###############################
### Multi-degree Splines
### lambda, tau
###############################

# 2021. 09. 14(TUE)

# rm(list = ls())

mds_cda = function(x, y, initial_coefficients = NULL, n_lambda = 50, n_tau = 10,
         lambda_max = NULL, lambda_max_min_ratio = 1e-4,
         iter_max = 1000, epsilon = 1e-4, verb = F,
         K = NULL, knots = NULL, degree = 1)
{
  # sample size
  N = length(x)
  # setting
  degrees = c(0, degree)
  basis1 = Psplines(x, K = K, knots = knots, degree = degrees[1]) # degree = 0
  basis2 = Psplines(x, K = K, knots = knots, degree = degrees[2]) # degree = d > 0
  dim1 = basis1$dimension
  dim2 = basis2$dimension
  knots = basis1$knots
  K = length(knots)
  
  B = cbind(basis2$basis[, 1 : (1 + degrees[2])],    # polynomial
            basis1$basis[, -1],                      # P_{0,k}
            basis2$basis[, -(1 : (1 + degrees[2]))]) # P_{d,k}
  p = ncol(B)
  
  r_theta_list = obj_f_list = aic_list = bic_list = matrix(0, n_lambda, n_tau)
  results = list()
  
  for(i in 1 : n_lambda)
  {
    results[[i]] = list()
    for(j in 1 : n_tau)
      results[[i]][[j]] = list()
  }
  
  # results = list(list())
  # results = list(results)
  
  # initial_coefficients
  if(is.null(initial_coefficients))
    initial_coefficients = rep(0, length = p)
  
  coefficients = initial_coefficients

  # lambda_max
  if(is.null(lambda_max))
  {
    lambda_tmp = c()
    for(j in (degree + 2) : p)
      lambda_tmp[j - (degree + 1)] = abs(mean(B[, j] * y))
    lambda_max = max(lambda_tmp)
  }
  
  # lambda sequence
  lambda_seq = exp(seq(log(lambda_max), log(lambda_max * lambda_max_min_ratio), length = n_lambda))
  lambda_seq[n_lambda] = 0
  
  # tau sequence
  tau_seq = seq(0, 1, length = n_tau)
  tau.5 = which(tau_seq > 0.5) # tau > 0.5 model index
  
  residuals = y - B %*% coefficients
  r_theta0 = mean(residuals^2) / 2
  obj_f0 = r_theta0
  step = 0 # model index
  
  # update start
  for(idx_lam in 1 : n_lambda)
  {
    lambda = lambda_seq[idx_lam]
    # results[[idx_lam]] = list()
    for(idx_tau in 1 : n_tau)
    {
      tau = tau_seq[idx_tau]
      for(iter in 1 : iter_max)
      {
        # alpha update
        # polynomial
        up_idx1 = 1 : (1 + degree)
        for(k1 in up_idx1)
        {
          partial_residuals = residuals + coefficients[k1] * B[, k1]
          coefficients[k1] = sum(partial_residuals * B[, k1]) / sum(B[, k1]^2)
          residuals = partial_residuals - coefficients[k1] * B[, k1]
        }
        # beta update
        # lasso penalty for degree[1] = 0
        up_idx2 = (2 + degree) : (2 + degree + K - 1)
        for(k2 in up_idx2)
        {
          partial_residuals = residuals + coefficients[k2] * B[, k2]
          coef_tmp = mean(partial_residuals * B[, k2])
          coefficients[k2] = soft_thresholding(coef_tmp, lambda * tau) / mean(B[, k2]^2)
          residuals = partial_residuals - coefficients[k2] * B[, k2]
        }
        # gamma update
        # lasso penalty for degree[2] > 0
        up_idx3 = (max(up_idx2) + 1) : p
        for(k3 in up_idx3)
        {
          partial_residuals = residuals + coefficients[k3] * B[, k3]
          coef_tmp = mean(partial_residuals * B[, k3])
          coefficients[k3] = soft_thresholding(coef_tmp, lambda * (1 - tau)) / mean(B[, k3]^2)
          residuals = partial_residuals - coefficients[k3] * B[, k3]
        }
        r_theta = mean(residuals^2) / 2
        
        # object function update
        obj_f = r_theta + lambda * (tau * sum(abs(coefficients[up_idx2])) + lambda * (1 - tau) * sum(abs(coefficients[up_idx3])))
        
        # stopping rule
        if(abs(obj_f - obj_f0) <= epsilon)
          break
        r_theta0 = r_theta
        obj_f0 = obj_f
      }# iteration
      step = step + 1
      if(verb == T)
        cat(step, "th model : ", iter, "th iteration, ", "r_theta = ", r_theta, "\n")
      
      fitted_value_degree1 = B[, c(1, up_idx2)] %*% coefficients[c(1, up_idx2)] # beta * x
      fitted_value_degree2 = B[, -up_idx2] %*% coefficients[-up_idx2]           # gamma * x
      
      k = sum(coefficients[-1] != 0)
      aic = N * log(2 * obj_f) + k * 2
      bic = N * log(2 * obj_f) + k * log(N)
      tune = paste("lambda =", round(lambda, 6), ", tau =", round(tau, 4), sep = " ")
      
      # matrix
      aic_list[idx_lam, idx_tau] = aic
      bic_list[idx_lam, idx_tau] = bic
      r_theta_list[idx_lam, idx_tau] = r_theta
      obj_f_list[idx_lam, idx_tau] = obj_f
      
      results[[idx_lam]][[idx_tau]]$coefficients          = coefficients
      results[[idx_lam]][[idx_tau]]$iteration             = iter
      results[[idx_lam]][[idx_tau]]$r_theta               = r_theta
      results[[idx_lam]][[idx_tau]]$obj_f                 = obj_f
      results[[idx_lam]][[idx_tau]]$aic                   = aic
      results[[idx_lam]][[idx_tau]]$bic                   = bic
      results[[idx_lam]][[idx_tau]]$tuning_parameters     = tune
      results[[idx_lam]][[idx_tau]]$fitted_values         = y - residuals # B %*% coefficients
      results[[idx_lam]][[idx_tau]]$fitted_value_degree1  = fitted_value_degree1
      results[[idx_lam]][[idx_tau]]$fitted_value_degree2  = fitted_value_degree2
    }# tau
  }# lambda
  
  # BIC matrix names
  rownames(bic_list) = 1 : n_lambda
  colnames(bic_list) = 1 : n_tau
  
  # all tau
  row = which(bic_list == min(bic_list), arr.ind = T)[1] # lambda index
  col = which(bic_list == min(bic_list), arr.ind = T)[2] # tau index
  
  # tau > 0.5
  # row.5 = which(bic_list[, tau.5] == min(bic_list[, tau.5]), arr.ind = T)[1]                  # lambda index
  # col.5 = which(bic_list[, tau.5] == min(bic_list[, tau.5]), arr.ind = T)[2] + length(!tau.5) # tau index
  
  results$lambda          = lambda_seq
  results$tau             = tau_seq
  results$B               = B
  results$knots           = knots
  results$aic             = aic_list
  results$bic             = bic_list
  results$min_bic_index   = paste("lambda index = ", row, ", tau index = ", col, sep = "")
  # results$min_bic_index.5 = paste("lambda index = ", row.5, ", tau index = ", col.5, sep = "")
  results$final_all_tau   = c(row, col)
  # results$final_tau.5     = c(row.5, col.5)
  
  results$best_bic = results[[row]][[col]]
  
  return(results)
}#multi-degree splines

