###########################
### QP
###########################


qp_sim = function(x, y, K, knots, degree, n_tau = 10, n_t = 50, t_max = NULL,
                  t_max_min_ratio = 1e-4, epsilon = 1e-10, verb = T)
{
  # setting
  basis1 = Psplines(x, K = K, knots = knots, degree = 0)      # degree = 0
  basis2 = Psplines(x, K = K, knots = knots, degree = degree) # degree = d > 0
  degrees = c(0, degree)
  n = length(x)
  aic_list = bic_list = matrix(0, n_t, n_tau)
  
  result = list()
  for(i in 1 : n_t)
  {
    result[[i]] = list()
    for(j in 1 : n_tau)
      result[[i]][[j]] = list()
  }
  
  # Design Matrix
  B = cbind(basis2$basis[, 1 : (1 + degree)],    # 1, polynomial
            basis1$basis[, -1],                  # P_{0,k}
            basis2$basis[, -(1 : (1 + degree))]) # P_{d,k}
  
  # tau sequence
  tau_seq = seq(0, 1, length = n_tau)
  
  # t max
  if(is.null(t_max))
  {
    lm.fit = lm(y ~ B - 1)
    beta   = na.omit(lm.fit$coefficients[(2 + degree) : (1 + degree + K)])
    gamma  = na.omit(lm.fit$coefficients[(2 + degree + K) : (1 + degree + 2 * K)])
    t_seq = matrix(0, n_t, n_tau)
    t_max = rep(0, n_tau)
    for(i in 1 : n_tau)
      t_max[i] = tau_seq[i] * sum(abs(beta)) + (1 - tau_seq[i]) * sum(abs(gamma))
  }
  
  # t sequence
  for(i in 1 : n_t)
  {
    for(j in 1 : n_tau)
    {
      # t_seq[, j] = exp(seq(log(t_max[j] * t_max_min_ratio), log(t_max[j]), length = n_t))
      t_seq[, j] = seq(t_max[j] * t_max_min_ratio, t_max[j], length = n_t)
    }
  }
  
  step1 = step2 = 0
  # QP start
  for(idx_t in 1 : n_t)
  {
    step1 = step1 + 1
    for(idx_tau in 1 : n_tau)
    {
      t = t_seq[idx_t, idx_tau]
      tau = tau_seq[idx_tau]
      # Quadratic Programming
      # Z
      Z = cbind(basis2$basis[, 1 : (1 + degree)],           # 1, polynomial
                basis1$basis[, -1],                         # P_{0}
                basis1$basis[, -1] * (-1),                  # -P_{0}
                basis2$basis[, -(1 : (1 + degree))],        # P_{d}
                basis2$basis[, -(1 : (1 + degree))] * (-1)) # -P_{d})
      # D
      Dmat = (t(Z) %*% Z) / n + diag(rep(epsilon, 1 + degree + 4 * K))
      # d
      dvec = t(y) %*% Z / n
      # b
      bvec = c(-t, rep(0, 4 * K))
      # Condition Matrix
      Amat = matrix(0, (1 + 4 * K), (1 + degree + 4 * K))
      Amat[1, -c(1 : (1 + degree))] = c(rep(-tau, 2 * K), rep((tau - 1), 2 * K))
      diag(Amat[-1, -c(1 : (1 + degree))]) = 1
      Amat = t(Amat)
      
      # QP solution
      library(quadprog)
      qp = solve.QP(Dmat = Dmat, dvec = dvec, Amat = Amat, bvec = bvec)
      
      step2 = step2 + 1
      if(verb == T)
        cat(step1, "th t,  ", step2, "th tau", "\n")
      
      tune = paste("t =", round(t, 6), ", tau =", round(tau, 6), sep = " ")
      
      alpha_hat    = qp$solution[1 : (1 + degree)]
      beta_plus    = qp$solution[(2 + degree) : (1 + degree + K)]
      beta_minus   = qp$solution[(2 + degree + K) : (1 + degree + 2 * K)]
      beta_hat     = beta_plus - beta_minus
      gamma_plus   = qp$solution[(2 + degree + 2 * K) : (1 + degree + 3 * K)]
      gamma_minus  = qp$solution[(2 + degree + 3 * K) : (1 + degree + 4 * K)]
      gamma_hat    = gamma_plus - gamma_minus
      coefficients = c(alpha_hat, beta_hat, gamma_hat)
      
      # Summary
      # Estimates
      result[[idx_t]][[idx_tau]]$alpha_hat    = alpha_hat
      result[[idx_t]][[idx_tau]]$beta_plus    = beta_plus
      result[[idx_t]][[idx_tau]]$beta_minus   = beta_minus
      result[[idx_t]][[idx_tau]]$beta_hat     = beta_plus - beta_minus
      result[[idx_t]][[idx_tau]]$gamma_plus   = gamma_plus
      result[[idx_t]][[idx_tau]]$gamma_minus  = gamma_minus
      result[[idx_t]][[idx_tau]]$gamma_hat    = gamma_hat
      result[[idx_t]][[idx_tau]]$coefficients = coefficients
      result[[idx_t]][[idx_tau]]$y_hat        = as.vector(B %*% coefficients)
      
      result[[idx_t]][[idx_tau]]$t     = t
      result[[idx_t]][[idx_tau]]$tau   = tau
      result[[idx_t]][[idx_tau]]$tune  = tune
      result[[idx_t]][[idx_tau]]$tuning_order = c(n_t, n_tau)
      
      # Model Selection Criteria
      residuals = y - as.vector(B %*% coefficients)
      obj_f = mean(residuals^2)
      k = sum(abs(coefficients[-1]) > 1e-8)
      aic = n * log(obj_f) + k * 2
      bic = n * log(obj_f) + k * log(n)
      
      result[[idx_t]][[idx_tau]]$aic = aic
      result[[idx_t]][[idx_tau]]$bic = bic
      
      aic_list[idx_t, idx_tau] = aic
      bic_list[idx_t, idx_tau] = bic
      
      result$degrees = degrees
      result$K       = K
      result$basis1  = basis1
      result$basis2  = basis2
      result$Z       = Z
      
      result$Dmat = Dmat
      result$dvec = dvec
      result$bvec = bvec
      result$Amat = Amat
      
      # result$solution               = qp$solution # coefficients
      # result$value                  = qp$value    # RSS_lam,tau
      # result$unconstrained.solution = qp$unconstrained.solution # LSE ?
      # result$iterations             = qp$iterations #
      # result$Lagrangian             = qp$Lagrangian # lambda ?
      # result$iact                   = qp$iact #
      
    } # tau
    step2 = 0
  } # t
  
  # BIC matrix names
  rownames(bic_list) = 1 : n_t   # t_seq
  colnames(bic_list) = 1 : n_tau # tau_seq
  
  # bic criteia
  row = which(bic_list == min(bic_list), arr.ind = T)[1] # t index
  col = which(bic_list == min(bic_list), arr.ind = T)[2] # tau index
  
  result$best_bic = result[[row]][[col]]
  
  result$aic = aic_list
  result$bic = bic_list
  return(result)
} # QP_simulation


# mds_qp = function(x, y, degree, tau, t, K = NULL, knots = NULL, epsilon = 1e-10)
# {
#   # setting
#   result = list()
#   degrees = c(0, degree)
#   basis1 = Psplines(x, K = K, knots = knots, degree = degrees[1]) # degree = 0
#   basis2 = Psplines(x, K = K, knots = knots, degree = degrees[2]) # degree = d > 0
#   
#   # the number of interior knots
#   if(is.null(K))
#     K = length(knots)
#   
#   # sample size
#   n = length(x)
#   
#   # Design Matrix
#   B = cbind(basis2$basis[, 1 : (1 + degrees[2])],    # 1, polynomial
#             basis1$basis[, -1],                      # P_{0,k}
#             basis2$basis[, -(1 : (1 + degrees[2]))]) # P_{d,k}
#   
#   # Quadratic Programming
#   Z = cbind(basis2$basis[, 1 : (1 + degrees[2])],           # 1, polynomial
#             basis1$basis[, -1],                             # P_{0}
#             basis1$basis[, -1] * (-1),                      # -P_{0}
#             basis2$basis[, -(1 : (1 + degrees[2]))],        # P_{d}
#             basis2$basis[, -(1 : (1 + degrees[2]))] * (-1)) # -P_{d}) 
#   
#   # D
#   Dmat = (t(Z) %*% Z) / n + diag(rep(epsilon, 1 + degree + 4 * K))
#   # d
#   dvec = t(y) %*% Z / n
#   # b
#   bvec = c(-t, rep(0, 4 * K))
#   # Condition Matrix
#   Amat = matrix(0, (1 + 4 * K), (1 + degree + 4 * K))
#   Amat[1, -c(1 : (1 + degree))] = c(rep(-tau, 2 * K), rep((tau - 1), 2 * K))
#   diag(Amat[-1, -c(1 : (1 + degree))]) = 1
#   Amat = t(Amat)
# 
#   qp = solve.QP(Dmat = Dmat, dvec = dvec, Amat = Amat, bvec = bvec)
#   
#   # Estimates
#   alpha_hat    = qp$solution[1 : (1 + degree)]
#   beta_plus    = qp$solution[(2 + degree) : (1 + degree + K)]
#   beta_minus   = qp$solution[(2 + degree + K) : (1 + degree + 2 * K)]
#   beta_hat     = beta_plus - beta_minus
#   gamma_plus   = qp$solution[(2 + degree + 2 * K) : (1 + degree + 3 * K)]
#   gamma_minus  = qp$solution[(2 + degree + 3 * K) : (1 + degree + 4 * K)]
#   gamma_hat    = gamma_plus - gamma_minus
#   coefficients  = c(alpha_hat, beta_hat, gamma_hat)
#   y_hat        = as.vector(B %*% coefficients)
#   
#   # Model Selection Criteria
#   rss = mean((y - y_hat)^2) / 2
#   obj_f = rss + tau * sum(abs(beta_hat)) + (1 - tau) * sum(abs(gamma_hat))
#   k = sum(coefficients[-1] != 0)
#   aic = n * log(2 * obj_f) + k * 2
#   bic = n * log(2 * obj_f) + k * log(n)
#   
#   tune = paste("t =", round(t, 6), ", tau =", round(tau, 4), sep = " ")
#   
#   # Summary
#   result$alpha         = alpha_hat
#   # result$beta_plus    = beta_plus
#   # result$beta_minus   = beta_minus
#   result$beta          = beta_hat
#   # result$gamma_plus   = gamma_plus
#   # result$gamma_minus  = gamma_minus
#   result$gamma         = gamma_hat
#   result$coefficients  = coefficients
#   result$y_hat         = y_hat
#   
#   result$t            = t
#   result$tau          = tau
#   result$tune         = tune
#   
#   result$degrees = degrees
#   result$K       = K
#   result$basis1  = basis1
#   result$basis2  = basis2
#   result$Z       = Z
#   
#   result$Dmat = Dmat
#   result$dvec = dvec
#   result$bvec = bvec
#   result$Amat = Amat
#   
#   # result$dim_D = dim(Dmat)
#   # result$dim_d = dim(dvec)
#   # result$dim_b = c(length(bvec), 1)
#   # result$dim_A = dim(Amat)
#   
#   result$solution               = qp$solution # coefficients
#   result$value                  = qp$value    # RSS_lam,tau
#   result$unconstrained.solution = qp$unconstrained.solution # LSE ?
#   result$iterations             = qp$iterations # 
#   result$Lagrangian             = qp$Lagrangian # lambda ?
#   result$iact                   = qp$iact #
#   
#   return(result)
# } # QP


###########################
### Plus Function
###########################
Plus = function(a)
{
  x_plus = a
  x_plus[a < 0] = 0
  return(x_plus)
}


########################################
### Truncated power splines Function 
########################################
# K : the number of interior knots
# knots : interior knots

Psplines = function(x, K = NULL, knots = NULL, degree)
{
  # sample size
  N = length(x)
  if(is.null(K))
    K = length(knots)
  else if(is.null(knots))
  {
    knots = seq(min(x), max(x), length.out = K + 2)
    knots = knots[-c(1, length(knots))]
  }else
    stop("Both K and knots are NULL!")
  
  X = matrix(0, N, K) # basis box
  
  # degree == 0
  if (degree == 0)
  {
    for(k in 1 : K)
    {
      X[, k] = Plus(x - knots[k])
      X[, k][X[, k] > 0] = 1
    }
    X = cbind(1, X)
    dimension = ncol(X)
    
    # basis colnames
    basis_name = c("intercept")
    for(k in 2 : dimension)
      basis_name = c(basis_name, paste0("(x-t_", k - 1, ")_+"))
    colnames(X) = basis_name
    
    #summary
    result = list()
    result$basis = X #matrix
    result$knots = knots
    result$dimension = ncol(X)
    result$degree = degree
    return(result)
  }
  
  # degree > 0
  for(k in 1 : K)
    X[, k] = Plus(x - knots[k])^degree # polynomial
  for(d in degree : 1)
    X = cbind(x^degree, X)             # psplines
  # intercept term
  X = cbind(1, X)
  # dimension
  dimension = ncol(X)
  # basis colnames
  basis_name = c("intercept")
  for(j in 2 : (1 + degree))
    basis_name = c(basis_name, paste0("x^", (j - 1)))
  for(k in 1 : length(knots))
    basis_name = c(basis_name, paste0("(x-t_", k, ")_+"))
  colnames(X) = basis_name
  
  # summary
  result = list()
  result$basis = X
  result$knots = knots
  result$dimension = dimension
  result$degree = degree
  return(result)
}
